@@ -3,10 +3,10 @@ from __future__ import annotations
 
 import asyncio
 import logging
-from datetime import datetime, time, timedelta
+from datetime import datetime, time, timedelta, timezone
 from enum import Enum, IntEnum, unique
 from json import dumps, loads
-from typing import TYPE_CHECKING, Any, cast
+from typing import TYPE_CHECKING, Any, Dict, Union, cast
 from uuid import uuid4
 
 from aiohttp import ClientWebSocketResponse, WSMsgType
@@ -19,7 +19,7 @@ except ImportError:  # pragma: no cover
 from ..activity import Activity, Insight
 from ..enums import LitterBoxStatus, LitterRobot4Command
 from ..exceptions import InvalidCommandException
-from ..utils import encode, utcnow
+from ..utils import encode, to_timestamp, utcnow
 from .litterrobot import LitterRobot
 from .models import LITTER_ROBOT_4_MODEL
 
@@ -39,15 +39,26 @@ LR4_STATUS_MAP = {
     "ROBOT_POWER_OFF": LitterBoxStatus.OFF,
     "ROBOT_POWER_UP": LitterBoxStatus.POWER_UP,
 }
+ACTIVITY_STATUS_MAP: dict[str, LitterBoxStatus | str] = {
+    "catDetectStuckLaser": LitterBoxStatus.CAT_SENSOR_FAULT,
+    "catWeight": "Pet Weight Recorded",
+    "powerTypeDC": "Battery Backup",
+    "robotCycleStateCatDetect": LitterBoxStatus.CAT_SENSOR_INTERRUPTED,
+    "robotCycleStatusDump": LitterBoxStatus.CLEAN_CYCLE,
+    "robotCycleStatusIdle": LitterBoxStatus.CLEAN_CYCLE_COMPLETE,
+    "robotStatusCatDetect": LitterBoxStatus.CAT_DETECTED,
+}
+
+LITTER_LEVEL_EMPTY = 500
 
 
 @unique
 class NightLightLevel(IntEnum):
     """Night light level of a Litter-Robot 4 unit."""
 
-    LOW = 85
-    MEDIUM = 170
-    HIGH = 255
+    LOW = 25
+    MEDIUM = 50
+    HIGH = 100
 
 
 @unique
@@ -62,6 +73,8 @@ class NightLightMode(Enum):
 class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
     """Data and methods for interacting with a Litter-Robot 4 automatic, self-cleaning litter box."""
 
+    _attr_model = "Litter-Robot 4"
+
     VALID_WAIT_TIMES = [3, 5, 7, 15, 30]
 
     _data_cycle_capacity = "DFINumberOfCycles"
@@ -81,6 +94,9 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
     _command_power_off = LitterRobot4Command.POWER_OFF
     _command_power_on = LitterRobot4Command.POWER_ON
 
+    _litter_level = LITTER_LEVEL_EMPTY
+    _litter_level_exp: datetime | None = None
+
     def __init__(self, data: dict, account: Account) -> None:
         """Initialize a Litter-Robot 4."""
         super().__init__(data, account)
@@ -108,6 +124,11 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
         """Return `True` if the drawer full indicator has been triggered."""
         return self._data.get("isDFIFull", False)
 
+    @property
+    def is_online(self) -> bool:
+        """Return `True` if the robot is online."""
+        return self._data.get("isOnline", False)
+
     @property
     def is_sleeping(self) -> bool:
         """Return `True` if the Litter-Robot is currently "sleeping" and won't automatically perform a clean cycle."""
@@ -130,13 +151,17 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
         ~ 461 low
         ~ 471 very low
         """
-        litter_level = int(self._data.get("litterLevel", 500))
-        return max(round(100 - (litter_level - 440) / 0.6, 1), 0)
-
-    @property
-    def model(self) -> str:
-        """Return the robot model."""
-        return "Litter-Robot 4"
+        new_level = int(self._data.get("litterLevel", LITTER_LEVEL_EMPTY))
+        now = datetime.now(timezone.utc)
+        if self._data.get("robotStatus") == "ROBOT_CLEAN":
+            self._litter_level_exp = now + timedelta(minutes=1)
+        elif (
+            self._litter_level_exp is None
+            or self._litter_level_exp < now
+            or abs(self._litter_level - new_level) < 10
+        ):
+            self._litter_level = new_level
+        return max(round(100 - (self._litter_level - 440) / 0.6, -1), 0)
 
     @property
     def night_light_brightness(self) -> int:
@@ -194,9 +219,10 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
     @property
     def status(self) -> LitterBoxStatus:
         """Return the status of the Litter-Robot."""
-        if self.is_waste_drawer_full:
+        status = self._data["robotStatus"]
+        if status == "ROBOT_IDLE" and self.is_waste_drawer_full:
             return LitterBoxStatus.DRAWER_FULL
-        return LR4_STATUS_MAP.get(self._data["robotStatus"], LitterBoxStatus.UNKNOWN)
+        return LR4_STATUS_MAP.get(status, LitterBoxStatus.UNKNOWN)
 
     @property
     def status_code(self) -> str | None:
@@ -221,6 +247,14 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
         ):
             self._parse_sleep_info()
 
+    def _parse_activity(self, activity: dict[str, str]) -> LitterBoxStatus | str:
+        """Parse an activity."""
+        value = activity["value"]
+        action = ACTIVITY_STATUS_MAP.get(value, value)
+        if value == "catWeight":
+            action = f"{action}: {activity['actionValue']} lbs"
+        return action
+
     def _parse_sleep_info(self) -> None:
         """Parse the sleep info of a Litter-Robot."""
         start = end = None
@@ -269,10 +303,11 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
                     "variables": {"serial": self.serial, "command": command, **kwargs},
                 }
             )
-            assert isinstance(data, dict)
-            if "Error" in (
-                error := data.get("data", {}).get("sendLitterRobot4Command", "")
-            ):
+            data = cast(dict, data)
+            error = data.get("data", {}).get("sendLitterRobot4Command", "")
+            if error and "Error" in error:
+                raise InvalidCommandException(error)
+            if error := ", ".join(e.get("message", "") for e in data.get("errors", {})):
                 raise InvalidCommandException(error)
             return True
         except InvalidCommandException as ex:
@@ -291,9 +326,35 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
                 "variables": {"serial": self.serial},
             },
         )
-        assert isinstance(data, dict)
+        data = cast(dict, data)
         self._update_data(data.get("data", {}).get("getLitterRobot4BySerial", {}))
 
+    async def set_name(self, name: str) -> bool:
+        """Set the name."""
+        data = await self._post(
+            json={
+                "query": """
+                    mutation rename(
+                        $serial: String!
+                        $name: String
+                    ) {
+                        updateLitterRobot4(
+                            input: {
+                                serial: $serial
+                                name: $name
+                            }
+                        ) {
+                            name
+                        }
+                    }
+                """,
+                "variables": {"serial": self.serial, "name": name},
+            }
+        )
+        updated_data = cast(dict, data).get("data", {}).get("updateLitterRobot4", {})
+        self._update_data(updated_data, partial=True)
+        return self.name == name
+
     async def set_night_light_brightness(
         self, brightness: int | NightLightLevel
     ) -> bool:
@@ -330,8 +391,53 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
 
     async def get_activity_history(self, limit: int = 100) -> list[Activity]:
         """Return the activity history."""
-        _LOGGER.warning("get_activity_history has not yet been implemented")
-        return []
+        if limit < 1:
+            raise InvalidCommandException(
+                f"Invalid range for parameter limit, value: {limit}, valid range: 1-inf"
+            )
+        data = await self._post(
+            json={
+                "query": """
+                    query GetLR4Activity(
+                        $serial: String!
+                        $startTimestamp: String
+                        $endTimestamp: String
+                        $limit: Int
+                        $consumer: String
+                    ) {
+                        getLitterRobot4Activity(
+                            serial: $serial
+                            startTimestamp: $startTimestamp
+                            endTimestamp: $endTimestamp
+                            limit: $limit
+                            consumer: $consumer
+                        ) {
+                            serial
+                            measure
+                            timestamp
+                            value
+                            actionValue
+                            originalHex
+                            valueString
+                            stateString
+                            consumer
+                            commandSource
+                        }
+                    }
+                """,
+                "variables": {
+                    "serial": self.serial,
+                    "limit": limit,
+                    "consumer": "app",
+                },
+            }
+        )
+        activities = cast(dict, data).get("data", {}).get("getLitterRobot4Activity", {})
+        return [
+            Activity(timestamp, self._parse_activity(activity))
+            for activity in activities
+            if (timestamp := to_timestamp(activity["timestamp"])) is not None
+        ]
 
     async def get_insight(
         self, days: int = 30, timezone_offset: int | None = None
@@ -340,26 +446,26 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
         data = await self._post(
             json={
                 "query": """
-                            query GetLR4Insights(
-                                $serial: String!
-                                $startTimestamp: String
-                                $timezoneOffset: Int
-                            ) {
-                                getLitterRobot4Insights(
-                                    serial: $serial
-                                    startTimestamp: $startTimestamp
-                                    timezoneOffset: $timezoneOffset
-                                ) {
-                                    totalCycles
-                                    averageCycles
-                                    cycleHistory {
-                                        date
-                                        numberOfCycles
-                                    }
-                                    totalCatDetections
-                                }
+                    query GetLR4Insights(
+                        $serial: String!
+                        $startTimestamp: String
+                        $timezoneOffset: Int
+                    ) {
+                        getLitterRobot4Insights(
+                            serial: $serial
+                            startTimestamp: $startTimestamp
+                            timezoneOffset: $timezoneOffset
+                        ) {
+                            totalCycles
+                            averageCycles
+                            cycleHistory {
+                                date
+                                numberOfCycles
                             }
-                        """,
+                            totalCatDetections
+                        }
+                    }
+                """,
                 "variables": {
                     "serial": self.serial,
                     "startTimestamp": (utcnow() - timedelta(days=days)).strftime(
@@ -374,14 +480,74 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
             insight["totalCycles"],
             insight["averageCycles"],
             [
-                Activity(
+                (
                     datetime.strptime(cycle["date"], "%Y-%m-%d").date(),
-                    count=cycle["numberOfCycles"],
+                    cycle["numberOfCycles"],
                 )
                 for cycle in insight["cycleHistory"]
             ],
         )
 
+    async def get_firmware_details(self) -> dict[str, bool | dict[str, str]]:
+        """Get the firmware details."""
+        data = await self._post(
+            json={
+                "query": """
+                    query getFirmwareDetails($serial: String!) {
+                        litterRobot4CompareFirmwareVersion(serial: $serial) {
+                            isEspFirmwareUpdateNeeded
+                            isPicFirmwareUpdateNeeded
+                            isLaserboardFirmwareUpdateNeeded
+                            latestFirmware {
+                                espFirmwareVersion
+                                picFirmwareVersion
+                                laserBoardFirmwareVersion
+                            }
+                        }
+                    }
+                """,
+                "variables": {"serial": self.serial},
+            }
+        )
+        data = cast(Dict[str, Dict[str, Dict[str, Union[bool, Dict[str, str]]]]], data)
+        return data.get("data", {}).get("litterRobot4CompareFirmwareVersion", {})
+
+    async def get_latest_firmware(self) -> str:
+        """Get the latest firmware available."""
+        latest_firmware = (await self.get_firmware_details()).get("latestFirmware", {})
+        latest_firmware = cast(Dict[str, str], latest_firmware)
+        return (
+            f"ESP: {latest_firmware.get('espFirmwareVersion')} / "
+            f"PIC: {latest_firmware.get('picFirmwareVersion')} / "
+            f"TOF: {latest_firmware.get('laserBoardFirmwareVersion')}"
+        )
+
+    async def has_firmware_update(self) -> bool:
+        """Check if a firmware update is available."""
+        firmware = await self.get_firmware_details()
+        return any(value for value in firmware.values() if isinstance(value, bool))
+
+    async def update_firmware(self) -> bool:
+        """Trigger a firmware update."""
+        data = await self._post(
+            json={
+                "query": """
+                    mutation updateFirmware($serial: String!) {
+                        litterRobot4TriggerFirmwareUpdate(input: {serial: $serial}) {
+                            isUpdateTriggered
+                            isEspFirmwareUpdateNeeded
+                            isPicFirmwareUpdateNeeded
+                            isLaserboardFirmwareUpdateNeeded
+                        }
+                    }
+                """,
+                "variables": {"serial": self.serial},
+            }
+        )
+        data = cast(dict, data)
+        firmware = data.get("data", {}).get("litterRobot4TriggerFirmwareUpdate", {})
+        return bool(firmware.get("isUpdateTriggered", False))
+
     async def subscribe_for_updates(self) -> None:
         """Open a web socket connection to receive updates."""
 
@@ -391,7 +557,8 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
             return await self._account.session.get_bearer_authorization()
 
         async def _subscribe(send_stop: bool = False) -> None:
-            assert self._ws
+            if not self._ws:
+                return
             if send_stop:
                 await self._ws.send_json(
                     {"id": self._ws_subscription_id, "type": "stop"}
@@ -463,7 +630,6 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
                     _LOGGER.debug("resubscribed")
 
         try:
-            assert self._account
             self._ws = await self._account.ws_connect(
                 f"{self._path}/realtime",
                 params={
@@ -488,6 +654,5 @@ class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
         if (websocket := self._ws) is not None and not websocket.closed:
             self._ws = None
             await websocket.send_json({"id": self._ws_subscription_id, "type": "stop"})
-            assert self._account
             await self._account.ws_disconnect(self.id)
             _LOGGER.debug("Unsubscribed from updates")
